/*  mkgradient v0.1 Copyright (C) 2005 Thomas S. <tom@eggdrop.ch>
 *  
 *  For more information visit: http://www.eggdrop.ch/projects/mkgradient/
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <png.h>
#include <stdlib.h>
#include <stdio.h>

/* #define DEBUG */

#define DIR_NONE 0
#define DIR_VERTICAL 1
#define DIR_HORIZONTAL 2

typedef struct gradient
{
  int s_r, s_g, s_b;
  int e_r, e_g, e_b;
  int direction;
  int length;
  char *output;
} gradient_t;

int hex2int(char byte)
{
  if (byte >= '0' && byte <= '9')
    return byte-'0';
  if (byte >= 'a' && byte <= 'f')
    return 10+(byte-'a');
  if (byte >= 'A' && byte <= 'F')
    return 10+(byte-'A');
  return -1;
}

/* returns 1 on success, 0 on error */
int readcolor(char *hex, int *r, int *g, int *b)
{
  int i;
  int a[6];

  if (strlen(hex) != 6)
    return 0;

  for (i=0;i<6;i++)
  {
    a[i] = hex2int(hex[i]);
    if (a[i] == -1) return 0;
  }

  *r = (a[0]*16)+a[1];
  *g = (a[2]*16)+a[3];
  *b = (a[4]*16)+a[5];
  
  return 1;
}

/* 0 = no error
 * 1 = couldn't open file
 * 2 = png_create_write_struct() failed
 * 3 = png_create_info_struct() failed
 * 4 = png_init_io() failed
 * 5 = couldn't write PNG header
 * 6 = couldn't write PNG data
 * 7 = couldn't write PNG end
 * this function doesn't check the arguments */
int mkgradient(gradient_t *g)
{
  FILE *f;
  png_byte *ptr;
  png_byte **vptr;
  png_structp pngptr;
  png_infop infoptr;
  png_byte **rows;
  unsigned char bytes[3];
  float steps[3];
  int i;
  int vert;

  if (g->direction == DIR_VERTICAL) vert = 1; else vert = 0;

  if (!vert)
    rows = malloc(sizeof(png_byte *));
  else
    rows = malloc(sizeof(png_byte *) * g->length);

  steps[0] = (float)(g->e_r-g->s_r)/g->length;
  steps[1] = (float)(g->e_g-g->s_g)/g->length;
  steps[2] = (float)(g->e_b-g->s_b)/g->length;

#ifdef DEBUG
  printf("Steps       : %f %f %f\n", steps[0], steps[1], steps[2]);
#endif
  
  f = fopen(g->output, "wb");
  if (!f) return 1;

  if (!vert)
  {
    rows[0] = (png_byte *)malloc(sizeof(png_byte) * g->length * 3);
    for (i = 0, ptr = rows[0] ; ptr < (rows[0] + g->length*3) ; ptr+=3, i++)
    {
      bytes[0] = g->s_r + (char)(steps[0]*i);
      bytes[1] = g->s_g + (char)(steps[1]*i);
      bytes[2] = g->s_b + (char)(steps[2]*i);
      memcpy(ptr, bytes, 3);
    }
  }
  else
  {
    for (i = 0, vptr = rows ; i < g->length ; vptr++, i++)
    {
      bytes[0] = g->s_r + (char)(steps[0]*i);
      bytes[1] = g->s_g + (char)(steps[1]*i);
      bytes[2] = g->s_b + (char)(steps[2]*i);
      rows[i] = (png_byte *)malloc(sizeof(png_byte) * 3);
      memcpy(*vptr, bytes, 3);
    }
  }
#ifdef DEBUG
  printf("Last bytes  : %d %d %d\n", bytes[0], bytes[1], bytes[2]);
#endif

  pngptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (!pngptr)
    return 2;
  
  infoptr = png_create_info_struct(pngptr); 
  if (!infoptr)
    return 3;

  if (setjmp(png_jmpbuf(pngptr)))
    return 4;

  png_init_io(pngptr, f);

  if (setjmp(png_jmpbuf(pngptr)))
    return 5;

  png_set_IHDR(pngptr, infoptr, vert?1:g->length, vert?g->length:1, 8, \
      PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_DEFAULT, \
      PNG_FILTER_TYPE_DEFAULT);

  png_write_info(pngptr, infoptr);

  if (setjmp(png_jmpbuf(pngptr)))
    return 6;

  png_write_image(pngptr, rows);

  if (setjmp(png_jmpbuf(pngptr)))
    return 7;

  png_write_end(pngptr, NULL);

  if (!vert)
    free(rows[0]);
  else
    for (i=0; i <g->length ; i++)
      free(rows[i]);
  free(rows);

  return 0;
}

int main(int argc, char *argv[])
{
  gradient_t g;
  int ret;
  
  if (argc != 6)
  {
    printf("Usage: %s <start color> <end color> "
           "<direction> <length> <file>\n\n", argv[0]);
    printf("Example: %s 000000 ffffff h 256 out.png\n", argv[0]);
    return 1;
  }

  if (!readcolor(argv[1], &g.s_r, &g.s_g, &g.s_b) || \
      !readcolor(argv[2], &g.e_r, &g.e_g, &g.e_b))
  {
    printf("Color format must be rrggbb\n");
    return 2;
  }

#ifdef DEBUG
  printf("Start color : %d %d %d\n", g.s_r, g.s_g, g.s_b);
  printf("End color   : %d %d %d\n", g.e_r, g.e_g, g.e_b);
#endif

  g.direction = DIR_NONE;
  if (argv[3][0] == 'v')
    g.direction = DIR_VERTICAL;
  if (argv[3][0] == 'h')
    g.direction = DIR_HORIZONTAL;

  if (g.direction == DIR_NONE)
  {
    printf("Direction must be 'h' (horizontal) or 'v' (vertical)\n");
    return 3;
  }

#ifdef DEBUG
  printf("Direction   : %s\n", \
      (g.direction==DIR_VERTICAL)?"vertical":"horizontal");
#endif

  g.length = atoi(argv[4]);

  if (g.length <= 0)
  {
    printf("Length must be greater than zero.\n");
    return 4;
  }

  g.output = argv[5];

#ifdef DEBUG
  printf("Length      : %d\n", g.length);
  printf("Output      : %s\n", g.output);
#endif

  if ((ret = mkgradient(&g)) != 0)
    printf("Error %d ocurred!\n", ret);
#ifdef DEBUG
  else
    printf("Written!\n");
#endif

  return 0;
}
